/* File: MethodTest.java
 * Contains: Main program for testing method selection and overloading
 * Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Created: January 1998
 * Updated: Fri Jan 16 01:48:58 1998 by Jeff Dalton
 * Copyright: (c) 1998, AIAI, University of Edinburgh
 */

package oplan.test;

import java.util.*;


/** A class for testing */

class MethodTest {

    public static void main(String[] argv) {

	/* 1st we establish that when Strings are used directly,
	   the static method for Strings is called, not the method
	   for Objects. */

	p("String vs Object:");

	p(typeString("apple")); // prints "String"

	/* But with a variable of type Object, the Object method is called. */

	Object s = "pie";
	p(typeString(s));	// prints "Object"

	/* The Object method is also called when we pass the String
	   through a function that's declared to return an Object. */

	p(typeString(identity("apple"))); // prints "Object"

	/* So much for String vs Object.  Now we switch to two new
	   classes that we define.  T extends Object, and T1 extends T.
	   We also stop using static methods. */

	/* When a T or T1 is used directly, the right (ie, most specific
	   applicable) method is called. */

	p("\nT vs T1:");

	p(new  T().tString());	// prints "T"
	p(new T1().tString());	// prints "T1"

	/* This continues to work when we use variables of type T. */

	T t = new T();
	T t1 = new T1();
	p( t.tString());	// prints "T"
	p(t1.tString());	// prints "T1"

	/* Things get trickier when we use Object variables. */

	Object oT = new T();
	Object oT1 = new T1();
	// p( oT.tString());	// not allowed: no Object tString method
	// p(oT1.tString());	// not allowed: no Object tString method

	/* But a cast to T still leaves the right method being called. *

	p(((T) oT).tString());	// prints "T"
	p(((T)oT1).tString());	// prints "T1"

	/* Now back to static methods. */

	/* All is well when we use objects directly. */

	p("\ntStr:");
	p(tStr(new T()));	// prints "T"
	p(tStr(new T1()));	// prints "T1"

	/* So we try variables and again find that it's the declared class
	   of the variable that matters, not the actual class of the object. */

	p(tStr(t));		// prints "T"
	p(tStr(t1));		// prints "T"
	p(tStr(oT));		// prints "Object"
	p(tStr(oT1));		// prints "Object"

	/* Is it really only static (ie, class) methods that work
	   so strangely?  No.  */

	p("\ntString2:");
	p(t.tString2(t));	// prints "T, T"
	p(t.tString2(t1));	// prints "T, T"
	p(t.tString2(oT));	// prints "T, Object"
	p(t.tString2(oT1));	// prints "T, Object"

	p(t1.tString2(t));	// prints "T1, T"
	p(t1.tString2(t1));	// prints "T1, T"
	p(t1.tString2(oT));	// prints "T1, Object"
	p(t1.tString2(oT1));	// prints "T1, Object"

	/* What if there's only a T1 method?  The first example works,
	   when objects are used directly.  We might expect t1.onlyT1(t)
	   to be rejected by the compiler, since there in fact isn't a
	   matching method.  But t1.onlyT1(t1) is also rejected,
	   and the message is "Method onlyT1(T) not found in class T".
	   This is strange.  For t1.tString2(t1), it was getting the
	   tString2 method from class T1, even though it was the
	   wrong tString2 method.  But now it doesn't seem to be
	   looking in class T1 at all.  It seems the compiler requires
	   that the textually apparent class must have a method,
	   even if some other method would actually be called. */

	p("\nOnly T1:");
	p(new T1().onlyT1(new T1()));  // prints "T1, only T1"

	// p(t1.onlyT1(t));	// compiler says no.
	// p(t1.onlyT1(t1));	// compiler says no.

	/* Let's test construtors.  (No, they don't work either.) */

	p("\nConstructors:");
	p(new T("apple").type);	// prints "String"
	p(new T(s).type);	// prints "Object"

	/* How about double dispatch?  There's a root class, in this
	   case T.  It and all subclasses have the following left method:

	     String left(T t) { return t.right(this); }

	   A two-arg operation F(T a, T b) gets implemented by a.left(b).
	   The left method then does the type-case for the left arg, a,
	   and the right method does the sub type-case for the right
	   arg, b.

	   For a.left(b), we assume that lookup works correctly w.r.t. a,
	   so the correct left method is found.  Then when the compiler
	   looks at t.right(this), it knows the actual class of this.
	   So it's no longer treating it as a T.  We assume that lookup
	   w.r.t. t in t.right(this) works correctly, and that lookup
	   w.r.t. this in r.right(this) works correctly when the
	   direct class of this is known, which it now is.  So ... we
	   should get the expected result.  And we do. */

	p("\nDouble lookup:");
	p(doubleLookup(t, t));	 // prints "T, T"
	p(doubleLookup(t, t1));  // prints "T, T1"
	p(doubleLookup(t1, t));	 // prints "T1, T"
	p(doubleLookup(t1, t1)); // prints "T1, T1"

    }

    static void p(String s) {	// just an abbreviation
	System.out.println(s);
    }

    static Object identity(Object a) {
	return a;
    }

    static String typeString(Object a) {
	return "Object";
    }

    static String typeString(String s) {
	return "String";
    }

    static String tStr(Object a) {
	return "Object";
    }

    static String tStr(T t) {
	return "T";
    }

    static String tStr(T1 t1) {
	return "T1";
    }

    static String doubleLookup(T a, T b) {
	return a.left(b);
    }

}


class T extends Object {

    String type = "no type";

    // We will want to test constructors

    T() {}			// why do we need this?

    T(Object a) {
	type = "Object";
    }

    T(String s) {
	type = "String";
    }

    String tString() {
	return "T";
    }

    String tString2(Object a) {
	 return "T, Object";
    }

    String tString2(T t) {
	return "T, T";
    }

    String tString2(T1 t1) {
	return "T, T1";
    }

    String left(T t) {
	return t.right(this);
    }

    String right(T t) {
	return "T, T";
    }

    String right(T1 t1) {
	return "T1, T";
    }

}

class T1 extends T {

    String tString() {
	return "T1";
    }

    String tString2(Object a) {
	 return "T1, Object";
    }

    String tString2(T t) {
	return "T1, T";
    }

    String tString2(T1 t1) {
	return "T1, T1";
    }

    String onlyT1(T1 t1) {
	 return "T1, only T1";
    }

    String left(T t) {
	return t.right(this);
    }

    String right(T t) {
	return "T, T1";
    }

    String right(T1 t1) {
	return "T1, T1";
    }

}

