/* File: Util.java
 * Contains: A class for useful static methods
 * Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Created: January 1998
 * Updated: Fri Jan 23 16:10:03 1998 by Jeff Dalton
 * Copyright: (c) 1998, AIAI, University of Edinburgh
 */

package oplan.util;

import java.util.*;
import java.io.*;


/** Class for useful static methods. */

public class Util {


    /*
     * Problem notification
     */


    /** Use this to tell the user about minor problems.  Warn prints
      a message to System.out, followed by a backtrace for the current
      thread. */

    public static void warn(String message) {
	System.err.println("\nWarning: " + message + "\n");
	Thread.dumpStack();
	System.err.println("");
    }


    /*
     * Simple, text-based user interaction.
     */

    public static String askLine(String prompt) {
	System.out.print(prompt + " ");
	System.out.flush();
	return Util.readLine(System.in);
    }


    /* 
     * Line reading.  In Java, we seem to have to write this ourself.
     */

    private static final int LINE_MAX = 1024;

    // What does final imply for modifiable objects?  /\/
    private static byte[] lineBuffer = new byte[LINE_MAX];

    /** Reads a line from an InputStream.  Java provides println
      for printing lines, but no similarly easy way to read them.
      This is provided as a static method, rather than as a class
      for line-reader objects, so that it can more easily be mixed
      with other input operations on the same stream. */

    // This version works only when the end of the line is also the
    // end of the input available, as it (usually?) is when we're
    // reading what the user types to System.in.

    public static String readLine(InputStream is) {
	int status;
	try { status = is.read(lineBuffer); }
	catch (IOException e) {
	    warn("IOException reading line from " + is);
	    return "";
	}
	if (status == -1)	// end of stream
	    return "";
	if (status == 0)	// nothing there?
	    return "";
	if (status == LINE_MAX) {
	    // Note that this might be a case where the line exactly fits.
	    // We should check whether the last character read is a newline.
	    warn("Line longer than " + LINE_MAX + " chars in " + is + ";\n"
		 + "Line begins: \""
		 + new String(lineBuffer, 0, 0, Math.min(20, LINE_MAX))
		 + "\".");
	    // At this point, we might want to read the rest of the line
	    // by calling readLine recursively and concatenate the result
	    // into one big string.
	}

	// Convert the buffer contents to a String, omitting the
	// newline at the end.
	return new String(lineBuffer, 0, 0, status-1);
    }


}
