package oplan.util.lisp;

import java.io.*;
import oplan.util.*;


/** A LispReader can be used to read Objects using a Lisp-like syntax.

  @see LispFileReader */

public class LispReader {

    protected LispTokenizer tk;

    /*
     * Constructors
     */

    public LispReader(LispTokenizer tk) {
	this.tk = tk;
    }

    public LispReader(InputStream is) {
	this.tk = new LispTokenizer(is);
    }

    // StringBufferInputStream supposedly does not correctly convert
    // characters to bytes (according to Java in a Nutshell) and in
    // any case has been deprecated.  But nothing is said against
    // ByteArrayInputStream.

    public LispReader(String s) {
	this.tk = new LispTokenizer(
		    new StringBufferInputStream(s));
    }

    /*
     * The reader
     */

    private static final Object CLOSE = new Object(); // close paren

    public Object readObject() {
	Object result = reader();
	if (result == CLOSE)
	    throw new LispReadException("Extra close paren");
	else
	    return result;
    }

    protected Object reader() {

	int ttype;

	// Get a token
	try { ttype = tk.nextToken(); }
	catch (IOException e) {
	    Debug.noteException(e);
	    throw new LispReadException("Tokenizer error");
	}

	// The tokenizer shouldn't return an EOL, because we don't
	// have that enabled.
	Debug.assert(!(ttype == tk.TT_EOL));

	// Various cases ...
	if (ttype == tk.TT_EOF) {
	    return Lisp.EOF;
	}
	else if (ttype == tk.TT_WORD) {
	    if (tk.sval.equals("nil"))
		return Lisp.NIL;
	    else
		return Symbol.intern(tk.sval);
	}
	else if (ttype == tk.TT_NUMBER) {
	    return new Double(tk.nval);
	}
	else if (ttype == (int)'(') {
	    return listreader();
	}
	else if (ttype == (int)')') {
	    return CLOSE;
	}
	else if (ttype == (int)'"') {
	    return tk.sval;	// a string
	}
	else {
	    // A char of no special meaning.
	    return Symbol.intern(String.valueOf((char)ttype));
	}

    }

    protected List listreader() {
	int ttype;
	Cons head = new Cons(Lisp.NIL, Lisp.NIL);
	Cons at = head;
	Object elt;

	while (((elt = reader()) != CLOSE) && (elt != Lisp.EOF)) {
	    at.cdr = new Cons(elt, Lisp.NIL);
	    at = (Cons)at.cdr;
	}
	if (elt == Lisp.EOF)
	    throw new LispReadException("EOF in list.");
	else
	    return head.cdr;
    }

}


class LispReadException extends RuntimeException {

    LispReadException(String message) {
	super(message);
    }

}
